from scipy.linalg.cython_blas cimport (
    ddot, dcopy, daxpy, dscal, dsyr, dsyrk, dnrm2, dgemm, dgemv, dsymv, dger,
    sdot, scopy, saxpy, sscal, ssyr, ssyrk, snrm2, sgemm, sgemv, ssymv, sger
)
from scipy.linalg.cython_lapack cimport (
    dlacpy, dposv, dpotrf, dpotrs, dgelsd,
    slacpy, sposv, spotrf, spotrs, sgelsd
)

ctypedef double (*ddot_)(const int*, const double*, const int*, const double*, const int*) noexcept nogil
ctypedef void (*dcopy_)(const int*, const double*, const int*, const double*, const int*) noexcept nogil
ctypedef void (*daxpy_)(const int*, const double*, const double*, const int*, double*, const int*) noexcept nogil
ctypedef void (*dscal_)(const int*, const double*, double*, const int*) noexcept nogil
ctypedef void (*dsyr_)(const char*, const int*, const double*, const double*, const int*, double*, const int*) noexcept nogil
ctypedef void (*dsyrk_)(const char*, const char*, const int*, const int*, const double*, const double*, const int*, const double*, double*, const int*) noexcept nogil
ctypedef double (*dnrm2_)(const int*, const double*, const int*) noexcept nogil
ctypedef void (*dgemm_)(const char*, const char*, const int*, const int*, const int*, const double*, const double*, const int*, const double*, const int*, const double*, double*, const int*) noexcept nogil
ctypedef void (*dgemv_)(const char*, const int*, const int*, const double*, const double*, const int*, const double*, const int*, const double*, double*, const int*) noexcept nogil
ctypedef void (*dsymv_)(const char*, const int*, const double*, const double*, const int*, const double*, const int*, const double*, double*, const int*) noexcept nogil
ctypedef void (*dger_)(const int*, const int*, const double*, const double*, const int*, const double*, const int*, double*, const int*) noexcept nogil

ctypedef void (*dposv__)(const char*, const int*, const int*, double*, const int*, double*, const int*, int*) noexcept nogil
ctypedef void (*dlacpy__)(const char*, const int*, const int*, const double*, const int*, double*, const int*) noexcept nogil
ctypedef void (*dpotrf__)(const char*, const int*, double*, const int*, int*) noexcept nogil
ctypedef void (*dpotrs__)(const char*, const int*, const int*, const double*, const int*, double*, const int*, int*) noexcept nogil
ctypedef void (*dgelsd__)(const int*, const int*, const int*,
             double*, const int*,
             double*, const int*,
             double*, const double*, int*, double*,
             const int*, int*, int*) noexcept nogil


ctypedef float (*sdot_)(const int*, const float*, const int*, const float*, const int*) noexcept nogil
ctypedef void (*scopy_)(const int*, const float*, const int*, const float*, const int*) noexcept nogil
ctypedef void (*saxpy_)(const int*, const float*, const float*, const int*, float*, const int*) noexcept nogil
ctypedef void (*sscal_)(const int*, const float*, float*, const int*) noexcept nogil
ctypedef void (*ssyr_)(const char*, const int*, const float*, const float*, const int*, float*, const int*) noexcept nogil
ctypedef void (*ssyrk_)(const char*, const char*, const int*, const int*, const float*, const float*, const int*, const float*, float*, const int*) noexcept nogil
ctypedef float (*snrm2_)(const int*, const float*, const int*) noexcept nogil
ctypedef void (*sgemm_)(const char*, const char*, const int*, const int*, const int*, const float*, const float*, const int*, const float*, const int*, const float*, float*, const int*) noexcept nogil
ctypedef void (*sgemv_)(const char*, const int*, const int*, const float*, const float*, const int*, const float*, const int*, const float*, float*, const int*) noexcept nogil
ctypedef void (*ssymv_)(const char*, const int*, const float*, const float*, const int*, const float*, const int*, const float*, float*, const int*) noexcept nogil
ctypedef float (*sger_)(const int*, const int*, const float*, const float*, const int*, const float*, const int*, float*, const int*) noexcept nogil

ctypedef void (*sposv__)(const char*, const int*, const int*, float*, const int*, float*, const int*, int*) noexcept nogil
ctypedef void (*slacpy__)(const char*, const int*, const int*, const float*, const int*, float*, const int*) noexcept nogil
ctypedef void (*spotrf__)(const char*, const int*, float*, const int*, int*) noexcept nogil
ctypedef void (*spotrs__)(const char*, const int*, const int*, const float*, const int*, float*, const int*, int*) noexcept nogil
ctypedef void (*sgelsd__)(const int*, const int*, const int*,
             float*, const int*,
             float*, const int*,
             float*, const float*, int*, float*,
             const int*, int*, int*) noexcept nogil

ctypedef enum CBLAS_ORDER:
    CblasRowMajor = 101
    CblasColMajor = 102

ctypedef CBLAS_ORDER CBLAS_LAYOUT

ctypedef enum cblas_TRANSPOSE:
    CblasNoTrans=111
    CblasTrans=112
    CblasConjTrans=113
    CblasConjNoTrans=114

ctypedef enum CBLAS_UPLO:
    CblasUpper=121
    CblasLower=122

ctypedef enum CBLAS_DIAG:
    CblasNonUnit=131
    CblasUnit=132

ctypedef enum CBLAS_SIDE:
    CblasLeft=141
    CblasRight=142

cdef public double cblas_ddot(const int n, const double  *x, const int incx, const double  *y, const int incy) noexcept nogil:
    return (<ddot_>ddot)(&n, x, &incx, y, &incy)

cdef public void cblas_dcopy(const int n, const double *x, const int incx, double *y, const int incy) noexcept nogil:
    (<dcopy_>dcopy)(&n, x, &incx, y, &incy)

cdef public void cblas_daxpy(const int n, const double alpha, const double *x, const int incx, double *y, const int incy) noexcept nogil:
    (<daxpy_>daxpy)(&n, &alpha, x, &incx, y, &incy)

cdef public void cblas_dscal(const int N, const double alpha, double *X, const int incX) noexcept nogil:
    (<dscal_>dscal)(&N, &alpha, X, &incX)

cdef public void cblas_dsyr(const int order, const int Uplo, const int N, const double alpha, const double *X, const int incX, double *A, const int lda) noexcept nogil:
    cdef char uplo = 0#'\0'
    if (order == CblasColMajor):
        if (Uplo == CblasLower):
            uplo = 76#'L'
        else:
            uplo = 85#'U'

    else:
        if (Uplo == CblasLower):
            uplo = 85#'U'
        else:
            uplo = 76#'L'
    (<dsyr_>dsyr)(&uplo, &N, &alpha, X, &incX, A, &lda)

cdef public void cblas_dsyrk(const int Order, const int Uplo, const int Trans,
         const int N, const int K, const double alpha, const double *A, const int lda, const double beta, double *C, const int ldc) noexcept nogil:
    cdef char uplo = 0#'\0'
    cdef char trans = 0#'\0'
    if (Order == CblasColMajor):
        if (Uplo == CblasUpper):
            uplo = 85#'U'
        else:
            uplo = 76#'L'

        if (Trans == CblasTrans):
            trans = 84#'T'
        elif (Trans == CblasConjTrans):
            trans = 67#'C'
        else:
            trans = 78#'N'

    else:
        if (Uplo == CblasUpper):
            uplo = 76#'L'
        else:
            uplo = 85#'U'

        if (Trans == CblasTrans):
            trans = 78#'N'
        elif (Trans == CblasConjTrans):
            trans = 78#'N'
        else:
            trans = 84#'T'

    (<dsyrk_>dsyrk)(&uplo, &trans, &N, &K, &alpha, A, &lda, &beta, C, &ldc)

cdef public double cblas_dnrm2 (const int N, const double  *X, const int incX) noexcept nogil:
    return (<dnrm2_>dnrm2)(&N, X, &incX)

cdef public void cblas_dgemm(const int Order, const int TransA, const int TransB, const int M, const int N, const int K,
         const double alpha, const double *A, const int lda, const double *B, const int ldb, const double beta, double *C, const int ldc) noexcept nogil:
    cdef char transA = 0#'\0'
    cdef char transB = 0#'\0'
    if (Order == CblasColMajor):
        if (TransA == CblasTrans):
            transA = 84#'T'
        elif (TransA == CblasConjTrans):
            transA = 67#'C'
        else:
            transA = 78#'N'

        if (TransB == CblasTrans):
            transB = 84#'T'
        elif (TransB == CblasConjTrans):
            transB = 67#'C'
        else:
            transB = 78#'N'

        (<dgemm_>dgemm)(&transA, &transB, &M, &N, &K, &alpha, A, &lda, B, &ldb, &beta, C, &ldc);

    else:
        if (TransA == CblasTrans):
            transB = 84#'T'
        elif (TransA == CblasConjTrans):
            transB = 67#'C'
        else:
            transB = 78#'N'

        if (TransB == CblasTrans):
            transA = 84#'T'
        elif (TransB == CblasConjTrans):
            transA = 67#'C'
        else:
            transA = 78#'N'

        (<dgemm_>dgemm)(&transA, &transB, &N, &M, &K, &alpha, B, &ldb, A, &lda, &beta, C, &ldc)

cdef public void cblas_dgemv(const int order,  const int TransA,  const int m, const int n,
         const double alpha, const double  *a, const int lda,  const double  *x, const int incx,  const double beta,  double  *y, const int incy) noexcept nogil:
    cdef char trans = 0#'\0'
    if (order == CblasColMajor):
        if (TransA == CblasNoTrans):
            trans = 78#'N'
        elif (TransA == CblasTrans):
            trans = 84#'T'
        else:
            trans = 67#'C'

        (<dgemv_>dgemv)(&trans, &m, &n, &alpha, a, &lda, x, &incx, &beta, y, &incy);

    else:
        if (TransA == CblasNoTrans):
            trans = 84#'T'
        elif (TransA == CblasTrans):
            trans = 78#'N'
        else:
            trans = 78#'N'

        (<dgemv_>dgemv)(&trans, &n, &m, &alpha, a, &lda, x, &incx, &beta, y, &incy);

cdef public void cblas_dsymv(const int order, const int Uplo, const int N, const double alpha, const double *A,
                 const int lda, const double *X, const int incX, const double beta, double *Y, const int incY) noexcept nogil:
    cdef char uplo = 0#'\0';
    if (order == CblasColMajor):
        if (Uplo == CblasUpper):
            uplo = 85#'U'
        else:
            uplo = 76#'L'

    else:
        if (Uplo == CblasUpper):
            uplo = 76#'L'
        else:
            uplo = 85#'U'

    (<dsymv_>dsymv)(&uplo, &N, &alpha, A, &lda, X, &incX, &beta, Y, &incY)

cdef public void cblas_dger(const int order, const int m, const int n, const double alpha,
                const double *x, const int incx, const double *y, const int incy, double *a, const int lda) noexcept nogil:
    if (order == CblasColMajor):
        (<dger_>dger)(&m, &n, &alpha, x, &incx, y, &incy, a, &lda);

    else:
        (<dger_>dger)(&n, &m, &alpha, y, &incy, x, &incx, a, &lda);

cdef public void dposv_(const char* uplo, const int* m, const int* n, double* x, const int* ldx, double* y, const int* ldy, int* info) noexcept nogil:
    (<dposv__>dposv)(uplo, m, n, x, ldx, y, ldy, info)

cdef public void dlacpy_(const char* uplo, const int* m, const int* n, const double* x, const int* ldx, double* y, const int* ldy) noexcept nogil:
    (<dlacpy__>dlacpy)(uplo, m, n, x, ldx, y, ldy)

cdef public void dpotrf_(const char* a1, const int* a2, double* a3, const int* a4, int* a5) noexcept nogil:
    (<dpotrf__>dpotrf)(a1, a2, a3, a4, a5)

cdef public void dpotrs_(const char* a1, const int* a2, const int* a3, const double* a4, const int* a5, double* a6, const int* a7, int* a8) noexcept nogil:
    (<dpotrs__>dpotrs)(a1, a2, a3, a4, a5, a6, a7, a8)

cdef public void dgelsd_(const int* a1, const int* a2, const int* a3,
             double* a4, const int* a5,
             double* a6, const int* a7,
             double* a8, const double* a9, int* a10, double* a11,
             const int* a12, int* a13, int* a14) noexcept nogil:
    (<dgelsd__>dgelsd)(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14)

########################

cdef public float cblas_sdot(const int n, const float  *x, const int incx, const float  *y, const int incy) noexcept nogil:
    return (<sdot_>sdot)(&n, x, &incx, y, &incy)

cdef public void cblas_scopy(const int n, const float *x, const int incx, float *y, const int incy) noexcept nogil:
    (<scopy_>scopy)(&n, x, &incx, y, &incy)

cdef public void cblas_saxpy(const int n, const float alpha, const float *x, const int incx, float *y, const int incy) noexcept nogil:
    (<saxpy_>saxpy)(&n, &alpha, x, &incx, y, &incy)

cdef public void cblas_sscal(const int N, const float alpha, float *X, const int incX) noexcept nogil:
    (<sscal_>sscal)(&N, &alpha, X, &incX)

cdef public void cblas_ssyr(const int order, const int Uplo, const int N, const float alpha, const float *X, const int incX, float *A, const int lda) noexcept nogil:
    cdef char uplo = 0#'\0'
    if (order == CblasColMajor):
        if (Uplo == CblasLower):
            uplo = 76#'L'
        else:
            uplo = 85#'U'

    else:
        if (Uplo == CblasLower):
            uplo = 85#'U'
        else:
            uplo = 76#'L'
    (<ssyr_>ssyr)(&uplo, &N, &alpha, X, &incX, A, &lda)

cdef public void cblas_ssyrk(const int Order, const int Uplo, const int Trans,
         const int N, const int K, const float alpha, const float *A, const int lda, const float beta, float *C, const int ldc) noexcept nogil:
    cdef char uplo = 0#'\0'
    cdef char trans = 0#'\0'
    if (Order == CblasColMajor):
        if (Uplo == CblasUpper):
            uplo = 85#'U'
        else:
            uplo = 76#'L'

        if (Trans == CblasTrans):
            trans = 84#'T'
        elif (Trans == CblasConjTrans):
            trans = 67#'C'
        else:
            trans = 78#'N'

    else:
        if (Uplo == CblasUpper):
            uplo = 76#'L'
        else:
            uplo = 85#'U'

        if (Trans == CblasTrans):
            trans = 78#'N'
        elif (Trans == CblasConjTrans):
            trans = 78#'N'
        else:
            trans = 84#'T'

    (<ssyrk_>ssyrk)(&uplo, &trans, &N, &K, &alpha, A, &lda, &beta, C, &ldc)

cdef public float cblas_snrm2 (const int N, const float  *X, const int incX) noexcept nogil:
    return (<snrm2_>snrm2)(&N, X, &incX)

cdef public void cblas_sgemm(const int Order, const int TransA, const int TransB, const int M, const int N, const int K,
         const float alpha, const float *A, const int lda, const float *B, const int ldb, const float beta, float *C, const int ldc) noexcept nogil:
    cdef char transA = 0#'\0'
    cdef char transB = 0#'\0'
    if (Order == CblasColMajor):
        if (TransA == CblasTrans):
            transA = 84#'T'
        elif (TransA == CblasConjTrans):
            transA = 67#'C'
        else:
            transA = 78#'N'

        if (TransB == CblasTrans):
            transB = 84#'T'
        elif (TransB == CblasConjTrans):
            transB = 67#'C'
        else:
            transB = 78#'N'

        (<sgemm_>sgemm)(&transA, &transB, &M, &N, &K, &alpha, A, &lda, B, &ldb, &beta, C, &ldc);

    else:
        if (TransA == CblasTrans):
            transB = 84#'T'
        elif (TransA == CblasConjTrans):
            transB = 67#'C'
        else:
            transB = 78#'N'

        if (TransB == CblasTrans):
            transA = 84#'T'
        elif (TransB == CblasConjTrans):
            transA = 67#'C'
        else:
            transA = 78#'N'

        (<sgemm_>sgemm)(&transA, &transB, &N, &M, &K, &alpha, B, &ldb, A, &lda, &beta, C, &ldc)

cdef public void cblas_sgemv(const int order,  const int TransA,  const int m, const int n,
         const float alpha, const float  *a, const int lda,  const float  *x, const int incx,  const float beta,  float  *y, const int incy) noexcept nogil:
    cdef char trans = 0#'\0'
    if (order == CblasColMajor):
        if (TransA == CblasNoTrans):
            trans = 78#'N'
        elif (TransA == CblasTrans):
            trans = 84#'T'
        else:
            trans = 67#'C'

        (<sgemv_>sgemv)(&trans, &m, &n, &alpha, a, &lda, x, &incx, &beta, y, &incy);

    else:
        if (TransA == CblasNoTrans):
            trans = 84#'T'
        elif (TransA == CblasTrans):
            trans = 78#'N'
        else:
            trans = 78#'N'

        (<sgemv_>sgemv)(&trans, &n, &m, &alpha, a, &lda, x, &incx, &beta, y, &incy);

cdef public void cblas_ssymv(const int order, const int Uplo, const int N, const float alpha, const float *A,
                 const int lda, const float *X, const int incX, const float beta, float *Y, const int incY) noexcept nogil:
    cdef char uplo = 0#'\0';
    if (order == CblasColMajor):
        if (Uplo == CblasUpper):
            uplo = 85#'U'
        else:
            uplo = 76#'L'

    else:
        if (Uplo == CblasUpper):
            uplo = 76#'L'
        else:
            uplo = 85#'U'

    (<ssymv_>ssymv)(&uplo, &N, &alpha, A, &lda, X, &incX, &beta, Y, &incY)

cdef public void cblas_sger(const int order, const int m, const int n, const float alpha,
                const float *x, const int incx, const float *y, const int incy, float *a, const int lda) noexcept nogil:
    if (order == CblasColMajor):
        (<sger_>sger)(&m, &n, &alpha, x, &incx, y, &incy, a, &lda);

    else:
        (<sger_>sger)(&n, &m, &alpha, y, &incy, x, &incx, a, &lda);

cdef public void sposv_(const char* uplo, const int* m, const int* n, float* x, const int* ldx, float* y, const int* ldy, int* info) noexcept nogil:
    (<sposv__>sposv)(uplo, m, n, x, ldx, y, ldy, info)

cdef public void slacpy_(const char* uplo, const int* m, const int* n, const float* x, const int* ldx, float* y, const int* ldy) noexcept nogil:
    (<slacpy__>slacpy)(uplo, m, n, x, ldx, y, ldy)

cdef public void spotrf_(const char* a1, const int* a2, float* a3, const int* a4, int* a5) noexcept nogil:
    (<spotrf__>spotrf)(a1, a2, a3, a4, a5)

cdef public void spotrs_(const char* a1, const int* a2, const int* a3, const float* a4, const int* a5, float* a6, const int* a7, int* a8) noexcept nogil:
    (<spotrs__>spotrs)(a1, a2, a3, a4, a5, a6, a7, a8)

cdef public void sgelsd_(const int* a1, const int* a2, const int* a3,
             float* a4, const int* a5,
             float* a6, const int* a7,
             float* a8, const float* a9, int* a10, float* a11,
             const int* a12, int* a13, int* a14) noexcept nogil:
    (<sgelsd__>sgelsd)(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14)




try:
    import setuptools
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension
import numpy as np
from Cython.Distutils import build_ext
import sys, os, subprocess, warnings, re


found_omp = True
def set_omp_false():
    global found_omp
    found_omp = False

## Modify this to pass your own BLAS/LAPACK linkage parameters:
custom_blas_link_args = []
custom_blas_compile_args = []
# example:
# custom_blas_link_args = ["-lopenblas"]
if len(custom_blas_link_args) or len(custom_blas_compile_args):
    build_ext_with_blas = build_ext

use_findblas = False
if not (len(custom_blas_link_args) or len(custom_blas_compile_args)):
    use_findblas = (("findblas" in sys.argv)
                     or ("-findblas" in sys.argv)
                     or ("--findblas" in sys.argv))
    if os.environ.get('USE_FINDBLAS') is not None:
        use_findblas = True
    if use_findblas:
        sys.argv = [a for a in sys.argv if a not in ("findblas", "-findblas", "--findblas")]
        from findblas.distutils import build_ext_with_blas
    else:
        build_ext_with_blas = build_ext

## Modify this to make the output of the compilation tests more verbose
silent_tests = not (("verbose" in sys.argv)
                    or ("-verbose" in sys.argv)
                    or ("--verbose" in sys.argv))

## Workaround for python<=3.9 on windows
try:
    EXIT_SUCCESS = os.EX_OK
except AttributeError:
    EXIT_SUCCESS = 0

## For debugging
if "--asan" in sys.argv:
    ADD_ASAN = True
    sys.argv.remove("--asan")
else:
    ADD_ASAN = False
if "--ggdb" in sys.argv:
    ADD_GGDB = True
    sys.argv.remove("--ggdb")
else:
    ADD_GGDB = False

class build_ext_subclass( build_ext_with_blas ):
    def build_extensions(self):
        is_windows = sys.platform[:3].lower() == "win"
        
        if self.compiler.compiler_type == 'msvc':
            for e in self.extensions:
                e.extra_compile_args += ['/O2', '/openmp', '/GL', '/fp:contract', '/fp:except-']
        else:
            if not self.check_for_variable_dont_set_march() and not self.check_cflags_contain_arch():
                self.add_march_native()
            self.add_openmp_linkage()
            self.add_no_math_errno()
            self.add_no_trapping_math()
            self.add_ffp_contract_fast()
            self.add_clang_fp_reassociate()
            self.add_O3()
            self.add_std_c99()
            self.add_link_time_optimization()

            ### Now add arguments as appropriate for good performance
            for e in self.extensions:

                if is_windows:
                    e.define_macros += [("NO_LONG_DOUBLE", None)]

            if ADD_ASAN:
                for e in self.extensions:
                    if self.compiler.compiler_type != "clang":
                        e.extra_compile_args += ["-fsanitize=address", "-static-libasan", "-ggdb"]
                    else:
                        e.extra_compile_args += ["-fsanitize=address", "-static-libsan", "-ggdb"]

            elif ADD_GGDB:
                for e in self.extensions:
                    e.extra_compile_args += ["-ggdb"]
                    e.define_macros += [("DEBUG", 1)]

                if self.compiler.compiler_type == "clang":
                    e.extra_compile_args += [
                        "-Wno-unknown-pragmas",
                        "-Wno-unknown-attributes",
                        "-Wno-pass-failed",
                    ]           

        ## If a custom BLAS/LAPACK is provided:
        if len(custom_blas_link_args) or len(custom_blas_compile_args):
            for e in self.extensions:
                e.extra_compile_args += custom_blas_compile_args
                e.extra_link_args += custom_blas_link_args
                e.define_macros = [m for m in e.define_macros if m[0] != "USE_FINDBLAS"]

        ## SYR in OpenBLAS is currently 10-15x slower than MKL, avoid it:
        ## https://github.com/xianyi/OpenBLAS/issues/3237
        for e in self.extensions:
            has_openblas_or_atlas = False
            if not has_openblas_or_atlas:
                for arg in e.extra_compile_args:
                    if (bool(re.search("openblas", str(arg).lower()))
                        or bool(re.search("atlas", str(arg).lower()))
                    ):
                        has_openblas_or_atlas = True
                        break
            if not has_openblas_or_atlas:
                for arg in e.extra_link_args:
                    if (bool(re.search("openblas", str(arg).lower()))
                        or bool(re.search("atlas", str(arg).lower()))
                    ):
                        has_openblas_or_atlas = True
                        break
            if not has_openblas_or_atlas:
                for arg in e.define_macros:
                    if (bool(re.search("openblas", str(arg[0]).lower()))
                        or bool(re.search("atlas", str(arg[0]).lower()))
                    ):
                        has_openblas_or_atlas = True
                        break
            if has_openblas_or_atlas:
                if "AVOID_BLAS_SYR" not in [m[0] for m in e.define_macros]:
                    e.define_macros += [("AVOID_BLAS_SYR", None)]
                e.define_macros = [macro for macro in e.define_macros if macro[0] != "USE_BLAS_SYR"]

        build_ext_with_blas.build_extensions(self)

    def check_cflags_contain_arch(self):
        if "CFLAGS" in os.environ:
            arch_list = [
                "-march", "-mcpu", "-mtune", "-msse", "-msse2", "-msse3",
                "-mssse3", "-msse4", "-msse4a", "-msse4.1", "-msse4.2",
                "-mavx", "-mavx2", "-mavx512"
            ]
            for flag in arch_list:
                if flag in os.environ["CFLAGS"]:
                    return True
        return False

    def check_for_variable_dont_set_march(self):
        return "DONT_SET_MARCH" in os.environ

    def add_march_native(self):
        args_march_native = ["-march=native", "-mcpu=native"]
        for arg_march_native in args_march_native:
            if self.test_supports_compile_arg(arg_march_native):
                for e in self.extensions:
                    e.extra_compile_args.append(arg_march_native)
                break

    def add_link_time_optimization(self):
        args_lto = ["-flto=auto", "-flto"]
        for arg_lto in args_lto:
            if self.test_supports_compile_arg(arg_lto):
                for e in self.extensions:
                    e.extra_compile_args.append(arg_lto)
                    e.extra_link_args.append(arg_lto)
                break

    def add_no_math_errno(self):
        arg_fnme = "-fno-math-errno"
        if self.test_supports_compile_arg(arg_fnme):
            for e in self.extensions:
                e.extra_compile_args.append(arg_fnme)
                e.extra_link_args.append(arg_fnme)

    def add_no_trapping_math(self):
        arg_fntm = "-fno-trapping-math"
        if self.test_supports_compile_arg(arg_fntm):
            for e in self.extensions:
                e.extra_compile_args.append(arg_fntm)
                e.extra_link_args.append(arg_fntm)

    def add_O3(self):
        arg_O3 = "-O3"
        if self.test_supports_compile_arg(arg_O3):
            for e in self.extensions:
                e.extra_compile_args.append(arg_O3)
                e.extra_link_args.append(arg_O3)

    def add_std_c99(self):
        arg_std_c99 = "-std=c99"
        if self.test_supports_compile_arg(arg_std_c99):
            for e in self.extensions:
                e.extra_compile_args.append(arg_std_c99)
                e.extra_link_args.append(arg_std_c99)

    def add_ffp_contract_fast(self):
        arg_ffpc = "-ffp-contract=fast"
        if self.test_supports_compile_arg(arg_ffpc):
            for e in self.extensions:
                e.extra_compile_args.append(arg_ffpc)
                e.extra_link_args.append(arg_ffpc)

    def add_clang_fp_reassociate(self):
        if self.test_supports_clang_reassociate():
            for e in self.extensions:
                e.define_macros.append(("CLANG_FP_REASSOCIATE", None))

    def add_openmp_linkage(self):
        arg_omp1 = "-fopenmp"
        arg_omp2 = "-fopenmp=libomp"
        args_omp3 = ["-fopenmp=libomp", "-lomp"]
        arg_omp4 = "-qopenmp"
        arg_omp5 = "-xopenmp"
        is_apple = sys.platform[:3].lower() == "dar"
        args_apple_omp = ["-Xclang", "-fopenmp", "-lomp"]
        args_apple_omp2 = ["-Xclang", "-fopenmp", "-L/usr/local/lib", "-lomp", "-I/usr/local/include"]
        has_brew_omp = False
        if is_apple:
            try:
                res_brew_pref = subprocess.run(["brew", "--prefix", "libomp"], capture_output=True)
                if res_brew_pref.returncode == EXIT_SUCCESS:
                    brew_omp_prefix = res_brew_pref.stdout.decode().strip()
                    args_apple_omp3 = ["-Xclang", "-fopenmp", f"-L{brew_omp_prefix}/lib", "-lomp", f"-I{brew_omp_prefix}/include"]
                    has_brew_omp = True
            except Exception as e:
                pass


        if self.test_supports_compile_arg(arg_omp1, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp1)
                e.extra_link_args.append(arg_omp1)
        elif is_apple and self.test_supports_compile_arg(args_apple_omp, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += ["-lomp"]
        elif is_apple and self.test_supports_compile_arg(args_apple_omp2, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += ["-L/usr/local/lib", "-lomp"]
                e.include_dirs += ["/usr/local/include"]
        elif is_apple and has_brew_omp and self.test_supports_compile_arg(args_apple_omp3, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += [f"-L{brew_omp_prefix}/lib", "-lomp"]
                e.include_dirs += [f"{brew_omp_prefix}/include"]
        elif self.test_supports_compile_arg(arg_omp2, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-fopenmp=libomp"]
                e.extra_link_args += ["-fopenmp"]
        elif self.test_supports_compile_arg(args_omp3, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-fopenmp=libomp"]
                e.extra_link_args += ["-fopenmp", "-lomp"]
        elif self.test_supports_compile_arg(arg_omp4, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp4)
                e.extra_link_args.append(arg_omp4)
        elif self.test_supports_compile_arg(arg_omp5, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp5)
                e.extra_link_args.append(arg_omp5)
        else:
            set_omp_false()

    def test_supports_compile_arg(self, comm, with_omp=False):
        is_supported = False
        try:
            if not hasattr(self.compiler, "compiler"):
                return False
            if not isinstance(comm, list):
                comm = [comm]
            print("--- Checking compiler support for option '%s'" % " ".join(comm))
            fname = "cmfrec_compiler_testing.c"
            with open(fname, "w") as ftest:
                ftest.write(u"int main(int argc, char**argv) {return 0;}\n")
            try:
                if not isinstance(self.compiler.compiler, list):
                    cmd = list(self.compiler.compiler)
                else:
                    cmd = self.compiler.compiler
            except Exception:
                cmd = self.compiler.compiler
            if with_omp:
                with open(fname, "w") as ftest:
                    ftest.write(u"#include <omp.h>\nint main(int argc, char**argv) {return 0;}\n")
            try:
                val = subprocess.run(cmd + comm + [fname], capture_output=silent_tests).returncode
                is_supported = (val == EXIT_SUCCESS)
            except Exception:
                is_supported = False
        except Exception:
            pass
        try:
            os.remove(fname)
        except Exception:
            pass
        return is_supported

    def test_supports_clang_reassociate(self):
        is_supported = False
        try:
            if not hasattr(self.compiler, "compiler"):
                return False
            print("--- Checking compiler support for option '%s'" % "#pragma clang fp reassociate(on)")
            fname = "cmfrec_compiler_testing.c"
            with open(fname, "w") as ftest:
                ftest.write(u"int main(int argc, char**argv) {return 0;}\n")
            try:
                if not isinstance(self.compiler.compiler, list):
                    cmd = list(self.compiler.compiler)
                else:
                    cmd = self.compiler.compiler
            except Exception:
                cmd = self.compiler.compiler

            with open(fname, "w") as ftest:
                ftest.write(u"""
                void fma_extra(double *a, double w, double *b, int n)
                {
                    #pragma clang fp reassociate(on)
                    for (int ix = 0; ix < n; ix++)
                        a[ix] += w * b[ix] * b[ix];
                }
                int main(int argc, char **argv)
                {
                    double a[] = {1.,2.,3.};
                    double b[] = {4.,5.,6.};
                    double w = 2.;
                    int n = 3;
                    fma_extra(a, w, b, n);

                    return 0;
                }\n
                """)
            try:
                val = subprocess.run(cmd + [fname], capture_output=silent_tests).returncode
                is_supported = (val == EXIT_SUCCESS)
            except Exception:
                is_supported = False
        except Exception:
            pass
        try:
            os.remove(fname)
        except Exception:
            pass
        return is_supported


force_openblas = (("openblas" in sys.argv)
                  or ("-openblas" in sys.argv)
                  or ("--openblas" in sys.argv))
if force_openblas:
    sys.argv = [a for a in sys.argv if a not in ("openblas", "-openblas", "--openblas")]
if os.environ.get('USE_OPENBLAS') is not None:
    force_openblas = True
if (force_openblas):
    custom_blas_link_args = ["-lopenblas"]
    from Cython.Distutils import build_ext
    build_ext_with_blas = build_ext

setup(
    name  = "cmfrec",
    packages = ["cmfrec"],
    version = '3.5.1-9',
    description = 'Collective matrix factorization',
    author = 'David Cortes',
    url = 'https://github.com/david-cortes/cmfrec',
    keywords = ['collaborative filtering', 'collective matrix factorization',
                'relational learning'],
    install_requires=[
        'cython',
        'numpy>=1.25',
        'scipy',
        'pandas',
        'findblas'
    ],
    cmdclass = {'build_ext': build_ext_subclass},
    ext_modules = [
        Extension("cmfrec.wrapper_double",
            sources=["cmfrec/cfuns_double.pyx" if use_findblas else "cmfrec/cfuns_double_plusblas.pyx",
                     "src/collective.c", "src/common.c",
                     "src/offsets.c", "src/helpers.c", "src/lbfgs.c",
                     "src/cblas_wrappers.c"],
            include_dirs=[np.get_include(), "src"],
            define_macros = [("_FOR_PYTHON", None),
                             ("USE_DOUBLE", None),
                             ("NDEBUG", None),
                             ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
                             ("USE_FINDBLAS" if use_findblas else "NO_FINDBLAS", None),
                             ("USE_BLAS_SYR" if use_findblas else "AVOID_BLAS_SYR", None)]
            ),
        Extension("cmfrec.wrapper_float",
            sources=["cmfrec/cfuns_float.pyx" if use_findblas else "cmfrec/cfuns_float_plusblas.pyx",
                     "src/collective.c", "src/common.c",
                     "src/offsets.c", "src/helpers.c", "src/lbfgs.c",
                     "src/cblas_wrappers.c"],
            include_dirs=[np.get_include(), "src"],
            define_macros = [("_FOR_PYTHON", None),
                             ("USE_FLOAT", None),
                             ("NDEBUG", None),
                             ("USE_FINDBLAS" if use_findblas else "NO_FINDBLAS", None),
                             ("USE_BLAS_SYR" if use_findblas else "AVOID_BLAS_SYR", None)]
            ),
        ]
)

if not found_omp:
    omp_msg  = "\n\n\nCould not detect OpenMP. Package will be built without multi-threading capabilities. "
    omp_msg += " To enable multi-threading, first install OpenMP"
    if (sys.platform[:3] == "dar"):
        omp_msg += " - for macOS: 'brew install libomp'\n"
    else:
        omp_msg += " modules for your compiler. "
    
    omp_msg += "Then reinstall this package from scratch: 'pip install --upgrade --no-deps --force-reinstall cmfrec'.\n"
    warnings.warn(omp_msg)
